/*
 * Copyright (C) 1995, jack
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

static char rcsid[]= "$Id: zs.c,v 1.12 1995/11/14 09:14:01 jack Exp $";

#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>

#include "config.h"
#include "keyin.h"
#include "rsio.h"
#include "updcrc.h"
#include "utils.h"
#include "zsend.h"
#include "zreceive.h"
#include "zmisc.h"
#include "zmodem.h"
#include "zwork.h"

extern int verbose;

#define ITOB(a,x) \
do \
  { \
    int xx = (x); \
    (a)[ZP0] = ((xx) & 0xff); \
    (a)[ZP1] = ((xx >>= 8) & 0xff); \
    (a)[ZP2] = ((xx >>= 8) & 0xff); \
    (a)[ZP3] = ((xx >>= 8) & 0xff); \
  } \
while (0)
#define VNNEWLINE() \
do \
  if (verbose >= 0 && isatty (STDOUT_FILENO)) \
    printf ("\n"); \
while (0)
#define UPDATEBAR(fsize,rsize,tssize,ttotal) \
do \
  if (verbose >= 0) \
    zbarprint ((fsize), (rsize), (tssize), (ttotal)); \
while (0)
#define VHPRINT(workp,type,format) \
do \
  if (verbose > 0) \
    zvhprint ((workp), (type), (format)); \
while (0)
#define VPRINTF(format) \
do \
  if (verbose > 0) \
    printf format; \
while (0)
#define VNPRINTF(format) \
do \
  if (verbose >= 0) \
    printf format; \
while (0)
#define Seek(fd,off,mod) lseek (fd, off, mod)
#define Read(fd,buf,len) read (fd, buf, len)
#define Write(fd,buf,len) write (fd, buf, len)
#define Close(fd) close (fd)
#define get_filebuf() (readbuf)
#define get_filebufsiz() (readbufsize)

#define OK (0)
#define ESC (0x1b)
#define MAXSUBPACKETS (4)
#define BUFFERED_SIZE (16384)

static void quit (int sig);
int zs (char *files[]);
int zsendfiles (char *files[]);
int getzrxinit (void);
int sendzsinit (void);
char *zitostr (int value, char *string, int radix);
int zmkinfo (char *info);
int zsendfile (void);
int zsendfdata (void);
int zreadbuf (void);
void zsendfin (void);
int getinsync (int flag);

static void (*Original_sigint) (int) = NULL;
static void (*Original_sigquit) (int) = NULL;
static void (*Original_sigterm) (int) = NULL;
static char *readbuf = NULL;
static int readbufsize = BUFFERED_SIZE;

static void
quit (int sig)
{
  zcancel ();
  RSend ();
  Keyflush ();
  if (readbuf != NULL)
    {
      free (readbuf);
      readbuf = NULL;
    }
  exit (1);
}

int
zs (char *files[])
{
  struct zwork *zworkp = &zwork;

  zworkp->usefcs32 = 0;
  zworkp->ctlesc = Zctlesc;
  zworkp->retry = 0;
  zworkp->attn[0] = '\0';
  zworkp->fname[0] = '\0';
  zworkp->fd = -1;
  zworkp->fpos = 0;

  Original_sigint = signal (SIGINT, quit);
#ifdef SIGQUIT
  Original_sigquit = signal (SIGQUIT, quit);
#endif
  Original_sigterm = signal (SIGTERM, quit);
  readbuf = xmalloc (readbufsize);

  /* Request ZRINIT */
  ITOB (zworkp->sbuf, 0);
  VHPRINT (zworkp, ZRQINIT, "Send: %s\n");
  zshhdr (4, ZRQINIT, zworkp->sbuf);

  /* Receive ZRINIT and send ZSINIT */
  if (getzrxinit () != OK)
    {
      zcancel ();
exit_err:
      (void) signal (SIGINT, Original_sigint);
#ifdef SIGQUIT
      (void) signal (SIGQUIT, Original_sigquit);
#endif
      (void) signal (SIGTERM, Original_sigterm);
      free (readbuf);
      readbuf = NULL;

      return ERROR;
    }

  /* Send all files */
  switch (zsendfiles (files))
    {
    case OK:
      /* Send ZFIN */
      zsendfin ();
      break;
    default:
      goto exit_err;
    }

  (void) signal (SIGINT, Original_sigint);
#ifdef SIGQUIT
  (void) signal (SIGQUIT, Original_sigquit);
#endif
  (void) signal (SIGTERM, Original_sigterm);
  free (readbuf);
  readbuf = NULL;

  return OK;
}

int
zsendfiles (char *files[])
{
  struct zwork *zworkp = &zwork;
  int fd;
  int status = OK;
  int i;

  for (i = 0; files[i] != NULL; i++)
    {
      /* Exist check */
      if ((fd = open (files[i], O_RDONLY|O_BINARY)) < 0)
	{
	  VNPRINTF (("Cannot open: %s\n", files[i]));
	  continue;
	}

      zworkp->retry = 0;
      strcpy (zworkp->fname, files[i]);
      zworkp->fd = fd;
      zworkp->fpos = 0;
      Seek (zworkp->fd, zworkp->fpos, SEEK_SET);

      /* Make ZFILE information */
      if ((zworkp->infolen = zmkinfo (zworkp->fileinfo)) < 0)
	{
	  VNPRINTF (("Cannot file information:%s\n", zworkp->fname));
	  Close (zworkp->fd);
	  zworkp->fd = -1;
	  continue;
	}

      /* Send file info and body */
      switch (zsendfile ())
	{
	case ZSKIP:
	case ZFERR:
	case OK:
	  status = OK;
	  break;
	default:
	  status = ERROR;
	  break;
	}

      Close (zworkp->fd);
      zworkp->fd = -1;

      if (status == ERROR)
	break;
    }

  return status;
}

int
getzrxinit (void)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;

  for (n = 10; --n >= 0; )
    {
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      switch (c = zgethdr (zworkp->rbuf))
	{
	case ZCHALLENGE:
	  VHPRINT (zworkp, ZCHALLENGE, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  ITOB (zworkp->sbuf, zworkp->rpos);
	  VHPRINT (zworkp, ZACK, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zshhdr (4, ZACK, zworkp->sbuf);
	  continue;
	case ZCOMMAND:
	  VHPRINT (zworkp, ZCOMMAND, "Recv: %s\n");
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZRQINIT, "Send: %s\n");
	  zshhdr (4, ZRQINIT, zworkp->sbuf);
	  continue;
	case ZRINIT:
	  VHPRINT (zworkp, ZRINIT, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zworkp->rflags = (zworkp->rbuf[ZF0] & 0xff);
	  zworkp->usefcs32 = (zworkp->rflags & CANFC32);
	  zworkp->ctlesc |= (zworkp->rflags & ESCCTL);
	  zworkp->blksiz = ZMAXSPLEN;
	  return (sendzsinit ());
	case ZCAN:
	case TIMEOUT:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  return ERROR;
	case ZRQINIT:
	  VHPRINT (zworkp, ZRQINIT, "Recv: %s\n");
	  if (zworkp->rbuf[ZF0] == ZCOMMAND)
	    continue;
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZNAK, "Send: %s\n");
	  zshhdr (4, ZNAK, zworkp->sbuf);
	  continue;
	}
    }
  return ERROR;
}

int
sendzsinit (void)
{
  struct zwork *zworkp = &zwork;
  int c;
  int errors;

  if (Zattn[0] == '\0' && (!zworkp->ctlesc || (zworkp->rflags & ESCCTL)))
    return OK;

  errors = 0;
  while (1)
    {
      ITOB (zworkp->sbuf, 0);
      if (zworkp->ctlesc)
	{
	  zworkp->sbuf[ZF0] |= TESCCTL;
	  VHPRINT (zworkp, ZSINIT, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zshhdr (4, ZSINIT, zworkp->sbuf);
	}
      else
	{
	  VHPRINT (zworkp, ZSINIT, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zsbhdr (4, ZSINIT, zworkp->sbuf);
	}
      zsdata (Zattn, ZATTNLEN, ZCRCW);

      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      c = zgethdr (zworkp->rbuf);
      switch (c)
	{
	case ZCAN:
	  VHPRINT (zworkp, ZCAN, "Recv: %s\n");
	  return ERROR;
	case ZACK:
	  VHPRINT (zworkp, ZACK, "Recv: %s\n");
	  return OK;
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  if (++errors > 19)
	    return ERROR;
	  continue;
	}
    }
}

char *
zitostr (int value, char *string, int radix)
{
  static char digits[] = "0123456789abcdef";
  char buf[(sizeof (int) * 8) + 1 + 1];
  char *p, *q;
  short sign;

  if (radix < 2 || radix > 16)
    {
      *string = '0';
      return string;
    }

  if (value < 0)
    {
      sign = 1;
      value = -value;
    }
  else
    sign = 0;

  p = buf + sizeof buf;
  *--p = '\0';
  do
    {
      *--p = digits[value % radix];
      value /= radix;
    }
  while (value > 0);

  /* Zero suppress */
  while (*p++ == '0')
    ;
  if (p[-1] == '\0' && p[-2] == '0')
    p--;
  p--;

  if (sign)
    *--p = '-';

  q = string;
  while ((*q++ = *p++))
    ;

  return string;
}

int
zmkinfo (char *info)
{
  struct zwork *zworkp = &zwork;
  struct stat statbuf;
  char *p, *top = info;

  /* File name */
  p = get_basename (zworkp->fname);
  while ((*info++ = *p++))
    ;
  info[-1] = '\0';

  /* File size */
  if (stat (zworkp->fname, &statbuf) != 0)
    return ERROR;
  zworkp->fsize = statbuf.st_size;
  p = zworkp->sdbuf;
  zitostr (zworkp->fsize, p, 10);
  while ((*info++ = *p++))
    ;
  info[-1] = ' ';

  /* File date */
  zworkp->ftime = statbuf.st_mtime;
  p = zworkp->sdbuf;
  zitostr (zworkp->ftime, p, 8);
  while ((*info++ = *p++))
    ;

  /* Data end */
  info[-1] = '\0';

  return (int)(info - top);
}

int
zsendfile (void)
{
  struct zwork *zworkp = &zwork;
  int c;
  short n;
  unsigned long crc = 0;
  long lastcrcrq = -1;

  for (n = 10; --n >= 0; )
    {
      if (Zresume)
	zworkp->sbuf[ZF0] = ZCRESUM;
      else
	zworkp->sbuf[ZF0] = Zconv;
      zworkp->sbuf[ZF1] = Zmanag;
      zworkp->sbuf[ZF2] = Ztrans;
      zworkp->sbuf[ZF3] = 0;
      VHPRINT (zworkp, ZFILE, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
      zsbhdr (4, ZFILE, zworkp->sbuf);
      zsdata (zworkp->fileinfo, zworkp->infolen, ZCRCW);
again:
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      c = zgethdr (zworkp->rbuf);
      switch (c)
	{
	case ZRINIT:
	  VHPRINT (zworkp, ZRINIT, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  while ((c = RSin ()) > 0)
	    if (c == ZPAD)
	      goto again;
	  continue;
	case ZCAN:
	case TIMEOUT:
	case ZABORT:
	case ZFIN:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  return ERROR;
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  continue;
	case ERROR:
	case ZNAK:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  continue;
	case ZCRC:
	  VHPRINT (zworkp, ZCRC, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (zworkp->rpos != lastcrcrq)
	    {
	      int count, rcount;
	      int bufsiz = get_filebufsiz ();
	      char *buf = get_filebuf ();

	      crc = 0xffffffffL;
	      count = 0;
	      Seek (zworkp->fd, 0, SEEK_SET);
	      while ((rcount = Read (zworkp->fd, buf, bufsiz)) > 0)
		{
		  count += rcount;
		  if (count < zworkp->rpos)
		    crc = mkcrc32 (buf, rcount, crc);
		  else if (count - (zworkp->rpos - 1) > 0)
		    {
		      crc = mkcrc32 (buf, rcount - (count - (zworkp->rpos - 1)), crc);
		      break;
		    }
		  else
		    break;
		}
	      Seek (zworkp->fd, zworkp->fpos, SEEK_SET);
	      crc = ~crc;
	      lastcrcrq = zworkp->rpos;
	    }
	  ITOB (zworkp->sbuf, crc);
	  VHPRINT (zworkp, ZCRC, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
	  zsbhdr (4, ZCRC, zworkp->sbuf);
	  goto again;
	case ZFERR:
	case ZSKIP:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  return c;
	case ZRPOS:
	  VHPRINT (zworkp, ZRPOS, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (Seek (zworkp->fd, zworkp->rpos, SEEK_SET) < 0)
	    return ERROR;
	  zworkp->fpos = zworkp->rpos;
	  return zsendfdata ();
	}
    }
  return ERROR;
}

int
zsendfdata (void)
{
  struct zwork *zworkp = &zwork;
  clock_t tstime, ttime, tdstime, tdtime, tdtotal;
  time_t tcomp;
  int tssize, tsize;
  short subpackets;
  int c;

  zworkp->lastsync = 0;
  zworkp->bcount = 0;
  zworkp->bpos = 0;
  zworkp->bufsiz = get_filebufsiz ();
  zworkp->buffer = get_filebuf ();
  zworkp->retry = 0;

  tssize = zworkp->fpos;
  tdtime = 0;
  tdtotal = 0;
  tstime = clock ();

  zreadbuf ();

  VNPRINTF (("Sending file: %s\n", zworkp->fname));
  subpackets = 0;
  c = 0;
somemore:
  if (0)
    {
waitack:
      subpackets = 0;
      c = getinsync (0);
gotack:
      switch (c)
	{
	default:
	case ZCAN:
	  UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize, tdtotal);
	  VNNEWLINE ();
	  return ERROR;
	case ZSKIP:
	  UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize, tdtotal);
	  VNNEWLINE ();
	  return c;
	case ZACK:
	case ZRPOS:
	  break;
	case ZRINIT:
	  UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize, tdtotal);
	  VNNEWLINE ();
	  return OK;
	}
    }

  while (RSinSense ())
    {
      switch (RSin ())
	{
	case CAN:
	case ZPAD:
	  c = getinsync (1);
	  goto gotack;
	case XOFF:
	case XOFF | 0x80:
	  RSin ();
	  break;
	}
    }

  ITOB (zworkp->sbuf, (zworkp->fpos + zworkp->bpos));
  VHPRINT (zworkp, ZDATA, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
  zsbhdr (4, ZDATA, zworkp->sbuf);

  tdstime = clock ();
  while (zworkp->bcount >= 0)
    {
      int e;
      short len;

#if 0
      if (subpackets >= (MAXSUBPACKETS - 1))
	e = ZCRCW;
      else
	e = ZCRCG;
#else
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      if (RSinSense ())
	{
	  e = ZCRCW;
	  do
	    switch (RSin ())
	      {
	      case CAN:
	      case ZPAD:
		c = getinsync (1);
		if (c == ZACK)
		  break;
		zsdata (zworkp->sbuf, 0, ZCRCE);
		tdtotal += tdtime;
		goto gotack;
	      case XOFF:
	      case XOFF | 0x80:
		RSin ();
		break;
	      }
	  while (RSinSense ());
	}
      else
	e = ZCRCG;
#endif

      len = zworkp->blksiz;
      if (zworkp->bpos + len > zworkp->bcount)
	{
	  if (zworkp->bcount == zworkp->bufsiz)
	    {
	      zworkp->fpos += zworkp->bpos;
	      if (zworkp->bpos != zworkp->bcount)
		if (Seek (zworkp->fd, zworkp->fpos, SEEK_SET) < 0)
		  {
		      UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize,
				 tdtotal);
		      VNNEWLINE ();
		    return ERROR;
		  }
	      zreadbuf ();

	      if (len > zworkp->bcount)
		{
		  len = zworkp->bcount;
		  e = ZCRCE;
		}
	    }
	  else
	    {
	      len = zworkp->bcount - zworkp->bpos;
	      e = ZCRCE;
	    }
	}

      UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize, tdtotal + tdtime);
      zsdata (zworkp->buffer + zworkp->bpos, len, e);
      tdtime = clock () - tdstime;
      zworkp->bpos += len;
      subpackets++;

      if (e == ZCRCW)
	{
	  tdtotal += tdtime;
	  goto waitack;
	}
      else if (e == ZCRCE)
	{
	  tdtotal += tdtime;
	  break;
	}
    }

  UPDATEBAR (zworkp->fsize, zworkp->fpos + zworkp->bpos, tssize, tdtotal);
  VNNEWLINE ();
  while (1)
    {
      ITOB (zworkp->sbuf, (zworkp->fpos + zworkp->bpos));
      VHPRINT (zworkp, ZEOF, "Send: %s %2.2x %2.2x %2.2x %2.2x\n");
      zsbhdr (4, ZEOF, zworkp->sbuf);
      switch (getinsync (0))
	{
	case ZACK:
	  continue;
	case ZRPOS:
	  goto somemore;
	case ZRINIT:
	  ttime = clock () - tstime;
	  tcomp = time (NULL);
	  if (ttime == 0)
	    ttime = 1;
	  tsize = (zworkp->fpos + zworkp->bpos) - tssize;
	  VNPRINTF (("%s: %d bytes %.2f sec. %.2f bytes/s\n", zworkp->fname, tsize,
		    ((double) ttime / (double) CLK_TCK),
		    ((double) (tsize * CLK_TCK) / (double) ttime)));
	  zlogout (1, zworkp->fname, tcomp, tsize, ttime);
	  return OK;
	case ZSKIP:
	  return c;
	default:
	  return ERROR;
	}
    }
}

int
zreadbuf (void)
{
  struct zwork *zworkp = &zwork;

  zworkp->bcount = Read (zworkp->fd, zworkp->buffer, zworkp->bufsiz);
  zworkp->bpos = 0;

  return zworkp->bcount;
}

void
zsendfin (void)
{
  struct zwork *zworkp = &zwork;
  int c;

  while (1)
    {
      ITOB (zworkp->sbuf, 0);
      VHPRINT (zworkp, ZFIN, "Send: %s\n");
      zshhdr (4, ZFIN, zworkp->sbuf);
      switch (c = zgethdr (zworkp->rbuf))
	{
	case ZFIN:
	  VHPRINT (zworkp, ZFIN, "Recv: %s\n");
	  RSout ('O');
	  RSout ('O');
	  RSflush ();
	  return;
	case ZCAN:
	case TIMEOUT:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  return;
	}
    }
}

int
getinsync (int flag)
{
  struct zwork *zworkp = &zwork;
  int c;

  while (1)
    {
      if (KeySense ())
	if (Keyin () == ESC)
	  raise (SIGTERM);

      c = zgethdr (zworkp->rbuf);
      switch (c)
	{
	case ZCAN:
	case ZABORT:
	case ZFIN:
	case TIMEOUT:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  return ERROR;
	case ZRPOS:
	  VHPRINT (zworkp, ZRPOS, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (zworkp->lastsync == zworkp->rpos)
	    {
	      if (++zworkp->retry > 12)
		return ERROR;
	      if (zworkp->retry > 4)
		if (zworkp->blksiz > 32)
		  zworkp->blksiz /= 2;
	    }

	  if ((zworkp->rpos < zworkp->fpos)
	      || (zworkp->rpos + zworkp->blksiz > zworkp->fpos + zworkp->bcount))
	    {
	      if (Seek (zworkp->fd, zworkp->rpos, SEEK_SET) < 0)
		return ERROR;
	      zworkp->fpos = zworkp->rpos;
	      zreadbuf ();
	    }

	  zworkp->lastsync = zworkp->rpos;
	  return c;
	case ZACK:
	  VHPRINT (zworkp, ZACK, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  if (flag || (zworkp->fpos + zworkp->bpos) == zworkp->rpos)
	    return ZACK;
	  continue;
	case ZRINIT:
	  VHPRINT (zworkp, ZRINIT, "Recv: %s %2.2x %2.2x %2.2x %2.2x\n");
	  return c;
	case ZSKIP:
	  VHPRINT (zworkp, ZSKIP, "Recv: %s\n");
	  return c;
	case ERROR:
	default:
	  VHPRINT (zworkp, c, "Recv: %s\n");
	  ITOB (zworkp->sbuf, 0);
	  VHPRINT (zworkp, ZNAK, "Send: %s\n");
	  zsbhdr (4, ZNAK, zworkp->sbuf);
	  continue;
	}
    }
}
